﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using libusbK;

namespace ReadISO
{
    class ReadISOStream
    {
        private object locker = new object();
        private object locker2 = new object();
        private bool cancelReadStream = false;
        public int maxPendingTransfers = 3;
        public int isoPacketsPerXfer = 128; // An ISO packet is the amount of data in a microframe = 3k.
        private ulong numBytesReceived = 0;
        private ulong numBytesReceivedExternallyReadable = 0;

        public bool Cancel
        {
            get
            {
                lock (locker)
                    return cancelReadStream;
            }
            set
            {
                lock (locker)
                    cancelReadStream = value;
            }
        }

        private bool numBytesReceivedReadOK = false;
        public bool GetNumBytesReceived(ref ulong numBytesReceivedArg)
        {
            if (numBytesReceivedReadOK)
            {
                numBytesReceivedArg = numBytesReceivedExternallyReadable;
                numBytesReceivedReadOK = false;
                return true;
            }
            else
            {
                return false;
            }
        }
        
        private void AddNumBytesReceived(ulong addend)
        {
            numBytesReceived += addend;
            if (!numBytesReceivedReadOK)
            {
                numBytesReceivedExternallyReadable = numBytesReceived;
                numBytesReceivedReadOK = true;
            }
        }

        private void ExtractData(IsoTransferItem iti)
        {
            KISO_PACKET kp;

            Debug.WriteLine("");
            for (int i = 0; i < iti.Iso.NumberOfPackets; i++)
            {
                if (iti.Iso.GetPacket(i, out kp))
                {
                    if (kp.Length > 0)
                    {
                        Debug.Write(string.Format("IsoPacket[{0}] length = {1}", i, kp.Length));
                        Debug.WriteLine(string.Format(", Offset = {0}", kp.Offset));
                    }
                }
                else
                {
                    Debug.WriteLine("Call to GetPacket failed.");
                }
            }
        }

        public void ReadStream(UsbK usb_arg, ref WINUSB_PIPE_INFORMATION pipeInfo) //, out byte[] buffer)
        {
            ReadIsoTransferQueue transferQueue;
            // Dummy value for isoTransferItem prevents "Unassigned variable" compile error.
            IsoTransferItem isoTransferItem = new IsoTransferItem(); 
            int transferred;
            int errorCode;

            transferQueue = new ReadIsoTransferQueue(usb_arg, ref pipeInfo, maxPendingTransfers, isoPacketsPerXfer);
            Cancel = false;
            // Transfer processing loop
            do
            {
                // While buffers exist in the Idle list, submit them. 
                while (transferQueue.Idle.Count > 0 && !Cancel)
                {
                    errorCode = transferQueue.SubmitNextRead();
                    if (errorCode != 0)
                    {
                        Debug.WriteLine(string.Format("IsoReadPipe failed. ErrorCode: {0:X8}h", errorCode));
                        goto Done;
                    }
                }

                if (transferQueue.Pending.Count == 0)
                {
                    // If no pending transfers, then we are shutting down.
                    Debug.WriteLine("Done!");
                    goto Done;
                }

                // Wait for the oldest transfer to complete.
                errorCode = transferQueue.WaitRead(out isoTransferItem, 2000, out transferred);
                if (errorCode != 0)
                {
                    Debug.WriteLine(string.Format("OvlPool.Wait failed. ErrorCode: {0:X8}h", errorCode));
                    //goto Done;
                }
                else if (transferred > 0)
                {
                    AddNumBytesReceived((ulong)transferred);
                    //if ( 0 == (numBytesReceived % 4096))
                    //    Debug.WriteLine((numBytesReceived/1024).ToString() + "k received.");

                    //// Extract data from IsoTransferItem buffer. Put in accumulator buffer.
                    //ExtractData(isoTransferItem);
                }
                else
                {
                    //Debug.WriteLine("WaitRead completed with 0 bytes transferred.");
                }
            } while (true);

        Done:
            transferQueue.Destroy();
            numBytesReceived = 0;
        } // End ReadStream();
    }
}
