﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using libusbK;

namespace LumaUSB_ns
{
    class ReadIsoOnce
    {
        public OvlK OvlPool;
        //public WINUSB_PIPE_INFORMATION PipeInfo;
        public UsbK Usb;
        public IsoTransferItem isoTransferItem;

        public bool Read(UsbK usb_arg, byte pipeID , out byte[] buffer)
        {
            bool success = false;
            int numOvlKs = 1;
            int numIsoPackets = 8; // There is 1 packet/microframe, 8 microframes/frame.
            int maxPacketSize = 3072;
            int bufferSize = numIsoPackets * maxPacketSize;
            int numBytesTransferred;
            Usb = usb_arg;
            OvlPool = new OvlK(Usb.Handle, numOvlKs, KOVL_POOL_FLAG.NONE);
            this.isoTransferItem = new IsoTransferItem();

            isoTransferItem.Buffer = new byte[bufferSize];
            // Initialize buffer with recognizable data to distinguish from new data.
            for (int i = 0; i < isoTransferItem.Buffer.Length; i++)
            {
                isoTransferItem.Buffer[i] = 0xEE;
            }
            isoTransferItem.BufferGC = GCHandle.Alloc(isoTransferItem.Buffer, GCHandleType.Pinned);

            isoTransferItem.Iso = new IsoK(numIsoPackets, 0);
            isoTransferItem.Iso.SetPackets(maxPacketSize);

            OvlPool.Acquire(out isoTransferItem.Ovl);

            // Prepare the OvlK handle to be submitted.
            OvlPool.ReUse(isoTransferItem.Ovl);

            //// Set the pipe timeout policy.
            //int[] pipeTimeoutMS = new[] { -1 };
            //int sizeOfInt = Marshal.SizeOf(typeof(int));
            //Usb.GetPipePolicy(
            //    pipeID,
            //    (int)PipePolicyType.PIPE_TRANSFER_TIMEOUT,
            //    ref sizeOfInt,
            //    pipeTimeoutMS);
            //Debug.WriteLine(string.Format("Initial timeout = {0}", pipeTimeoutMS[0]));

            //pipeTimeoutMS[0] = 2000;
            //Usb.SetPipePolicy(
            //                  pipeID,
            //                  (int)PipePolicyType.PIPE_TRANSFER_TIMEOUT,
            //                  Marshal.SizeOf(typeof(int)),
            //                  pipeTimeoutMS);

            //pipeTimeoutMS[0] = -2;
            //sizeOfInt = Marshal.SizeOf(typeof(int));
            //Usb.GetPipePolicy(
            //    pipeID,
            //    (int)PipePolicyType.PIPE_TRANSFER_TIMEOUT,
            //    ref sizeOfInt,
            //    pipeTimeoutMS);
            //Debug.WriteLine(string.Format("Final timeout = {0}", pipeTimeoutMS[0]));

            // The data buffer was pinned earlier when it was allocated.  Always pin managed memory before using it
            // in an asynchronous function to keep the framework from tampering with it.
            Debug.WriteLine("Calling Usb.IsoReadPipe...");
            Debug.WriteLine(
                string.Format("NumberOfPackets = {0}", isoTransferItem.Iso.NumberOfPackets));
            Usb.IsoReadPipe(pipeID,
                            isoTransferItem.Buffer,
                            bufferSize,
                            isoTransferItem.Ovl,
                            isoTransferItem.Iso.Handle);
            Debug.WriteLine(
                string.Format("NumberOfPackets = {0}", isoTransferItem.Iso.NumberOfPackets));

            int errorCode = Marshal.GetLastWin32Error();
            Debug.WriteLine(string.Format("Usb.IsoReadPipe caused errorcode = {0}.", errorCode));
            // 997 is ERROR_IO_PENDING. For async, this means so far so good.
            // IE: The transfer is submitted but we won't know if a problem occurs until it is completed.
            if ((errorCode == ErrorCodes.IoPending) || (errorCode == ErrorCodes.Success))
            {
                success = true;
            }

            if (true == success)
            {
                // Wait for i/o request to complete.
                Debug.WriteLine("Calling OvlPool.Wait...");
                success = OvlPool.Wait(isoTransferItem.Ovl, 5000, KOVL_WAIT_FLAG.NONE, out numBytesTransferred);
                Debug.WriteLine("Returned from OvlPool.Wait. Success = " + success.ToString());
                Debug.WriteLine("numBytesTransferred = " + numBytesTransferred.ToString());
                Debug.WriteLine("NumberOfPackets = " + isoTransferItem.Iso.NumberOfPackets.ToString());

                KISO_PACKET kp;
                for (int i = 0; i < isoTransferItem.Iso.NumberOfPackets; i++)
                {
                    if (isoTransferItem.Iso.GetPacket(i, out kp))
                    {
                        Debug.WriteLine(string.Format("IsoPacket[{0}] length = {1}", i, kp.Length));
                    }
                    else
                    {
                        Debug.WriteLine("Call to GetPacket failed.");
                    }
                }
            }
            buffer = isoTransferItem.Buffer;
            return success;
        }

            //        bool success = OvlPool.Wait(isoTransferItem.Ovl, 1000, KOVL_WAIT_FLAG.NONE, out transferred);
            //if (!success)
            //{
            //    return Marshal.GetLastWin32Error();
            //}

    }
}
